from daft.daft import PyRecordBatch

DAFT_DASHBOARD_ENV_NAME: str

class ConnectionHandle:
    def shutdown(self, noop_if_shutdown: bool) -> None: ...
    def get_port(self) -> int: ...

def launch(noop_if_initialized: bool = False, port: int | None = None) -> ConnectionHandle: ...
def register_dataframe_for_display(record_batch: PyRecordBatch) -> str: ...
def generate_interactive_html(df_id: str) -> str: ...
